/* Tables*/

.data_type {
    font-size: 0.8em;
    font-weight: normal;
}

.column_name {
    font-size: 1.2em;
}

.info_items {
    margin: 10px 0;
    font-family: sans-serif; 
    font-size: 10px;
}

.optimus_table {
    font-size: 12px;
    margin: 0;
}

.optimus_table_container {
    display: inline-flex;
    flex-direction: column;
}

.optimus_table thead {
    font-family: sans-serif;
    border-bottom: 1px solid currentColor;
}

.optimus_table tbody {
    font-family: monospace;
}

.optimus_table tbody div {
    min-height:14px
}

.optimus_table_border {
    background-color: currentColor; opacity: 0.2; height: 1px; width: 100%
}

.highlight-col {
    background-color: rgba(240,240,0,0.18);
}

/* Profiler */

.main {
    width:100%;
    overflow:auto;
    border-bottom:1px solid #eeeeee;
    padding: 10px 0;
}
.panel_profiler {
    margin-right:2%;
    float:left;
    padding-bottom:2%;
}
.panel_profiler {
    font-family:monospace;
}
.title_profiler {
    padding:20px;
    background-color: #eeeeee
}
.info {
    overflow: auto
}
.main td, main th {
    padding:0em
}
.panel_profiler td {
    padding:0.2em
}
