CLUSTER_COL = "cluster"
COUNT_COL = "count"
RECOMMENDED_COL = "recommended"
CLUSTER_COUNT_COL = "cluster_count"
CLUSTER_SUM_COL = "cluster_sum"

NGRAM_COL = "ngram"
FINGERPRINT_COL = "fingerprint"
NGRAM_FINGERPRINT_COL = "ngram_fingerprint"

LEVENSHTEIN_DISTANCE = "LEVENSHTEIN_DISTANCE"
STRING_TO_INDEX = "string_to_index"
INDEX_TO_STRING = "index_to_string"
